const tp = require('@tabletop-playground/api');


// Ability and skill info
const abilities = {
    "STR": "Strength", 
    "DEX": "Dexterity", 
    "CON": "Constitution", 
    "INT": "Intelligence", 
    "WIS": "Wisdom", 
    "CHA": "Charisma"};
const skills = {
    "Acrobatics": "DEX",
    "Animal Handling": "WIS",
    "Arcana": "INT",
    "Athletics": "STR",
    "Deception": "CHA",
    "History": "INT",
    "Insight": "WIS",
    "Intimidation": "CHA",
    "Investigation": "INT",
    "Medicine": "WIS",
    "Nature": "INT",
    "Perception": "WIS",
    "Performance": "CHA",
    "Persuasion": "CHA",
    "Religion": "INT",
    "Sleight of Hand": "DEX",
    "Stealth": "DEX",
    "Survival": "WIS"
}


// Widgets that can be modified by the player
var nameTextBox = new tp.TextBox().setMaxLength(100).setFontSize(20).setText("Character Name");
var classTextBox = new tp.TextBox().setMaxLength(30).setText("Class");
var raceTextBox = new tp.TextBox().setMaxLength(30).setText("Race");
var backgroundTextBox = new tp.TextBox().setMaxLength(30).setText("Background");
var alignmentSelection = new tp.SelectionBox();
var xpTextBox = new tp.TextBox().setInputType(4).setMaxLength(7).setText("0");
var abilityTextBoxes = {};
var proficiencyTextBox;
var acTextBox;
var initiativeTextBox;
var inspirationTextBox;
var currentHPTextBox;
var maxHPTextBox;
var deathSuccessCheckBoxes = [];
var deathFailureCheckBoxes = [];
var saveCheckBoxes = [];
var skillCheckBoxes = [];

var canvas = new tp.Canvas();


// Helper functions
function scoreToModifier(scoreText) {
    return Math.floor((parseInt(scoreText) - 10) / 2);
}

function modifierToString(modifier) {
    if (modifier < 0) {
        return modifier.toString();
    }
    return "+" + modifier.toString();
}

function scoreToModifierString(scoreText) {
    return modifierToString(scoreToModifier(scoreText));
}

function getAbilityModifier(ability) {
    return scoreToModifier(abilityTextBoxes[ability].getText());
}

function addCounter(x, y, digits = 2, min = 0) {
    var plusButton = new tp.Button().setFontSize(9).setText("+");
    var minusButton = new tp.Button().setFontSize(9).setText("-");
    var textBox = new tp.TextBox().setInputType(4).setMaxLength(digits).setFontSize(19);
    if (min < 0) {  
        textBox.setInputType(3);
    }
    else {
        textBox.setInputType(4);
    }

    plusButton.onClicked.add(() => {
        textBox.setText(Math.min(10**digits - 1, parseInt(textBox.getText()) + 1).toString()); 
    });
    minusButton.onClicked.add(() => { 
        textBox.setText(Math.max(min, parseInt(textBox.getText()) - 1).toString());
    });

    canvas.addChild(plusButton, x, y, 20, 20);
    canvas.addChild(minusButton, x, y + 22, 20, 20);
    canvas.addChild(textBox, x + 22, y, digits * 15 + 9, 42);

    return textBox;
}


// Save / Load
function toJSON() {
    // Create save objects with abbreviated names to keep size small
    saveData = {
        "nm": nameTextBox.getText(),
        "cl": classTextBox.getText(),
        "rc": raceTextBox.getText(),
        "bg": backgroundTextBox.getText(),
        "al": alignmentSelection.getSelectedIndex(),
        "xp": parseInt(xpTextBox.getText()),
        "pf": parseInt(proficiencyTextBox.getText()),
        "ac": parseInt(acTextBox.getText()),
        "ini": parseInt(initiativeTextBox.getText()),
        "ins": parseInt(inspirationTextBox.getText()),
        "hp": parseInt(currentHPTextBox.getText()),
        "mhp": parseInt(maxHPTextBox.getText()),
        "ds": [],
        "df": [],
        "sv": [],
        "sk": []
    }

    for (ability in abilities) {
        saveData[ability] = parseInt(abilityTextBoxes[ability].getText());
    }

    for (checkBox of deathSuccessCheckBoxes) {
        saveData["ds"].push(checkBox.isChecked() ? 1 : 0);
    }
    for (checkBox of deathFailureCheckBoxes) {
        saveData["df"].push(checkBox.isChecked() ? 1 : 0);
    }
    for (checkBox of saveCheckBoxes) {
        saveData["sv"].push(checkBox.isChecked() ? 1 : 0);
    }
    for (checkBox of skillCheckBoxes) {
        saveData["sk"].push(checkBox.isChecked() ? 1 : 0);
    }
    
    return JSON.stringify(saveData);
}

function fromJSON(text) {
    try {
        var saveData = JSON.parse(text);
    }
    catch (e) {
        // Don't load anything if JSON can't be parsed
        return;
    }

    nameTextBox.setText(saveData.nm);
    classTextBox.setText(saveData.cl);
    raceTextBox.setText(saveData.rc);
    backgroundTextBox.setText(saveData.bg);
    alignmentSelection.setSelectedIndex(saveData.al);
    xpTextBox.setText(saveData.xp.toString());
    proficiencyTextBox.setText(saveData.pf.toString());
    acTextBox.setText(saveData.ac.toString());
    initiativeTextBox.setText(saveData.ini.toString());
    inspirationTextBox.setText(saveData.ins.toString());
    currentHPTextBox.setText(saveData.hp.toString());
    maxHPTextBox.setText(saveData.mhp.toString());

    for (ability in abilities) {
        abilityTextBoxes[ability].setText(saveData[ability].toString());
    }

    for (var i = 0; i < 3; i += 1) {
        deathSuccessCheckBoxes[i].setIsChecked(saveData.ds[i], true);
        deathFailureCheckBoxes[i].setIsChecked(saveData.df[i], true);
    }
    for (var i = 0; i < saveCheckBoxes.length; i += 1) {
        saveCheckBoxes[i].setIsChecked(saveData.sv[i], true);
    }
    for (var i = 0; i < skillCheckBoxes.length; i += 1) {
        skillCheckBoxes[i].setIsChecked(saveData.sk[i], true);
    }
}

// Store reference object in variable to use in closures to save and load
var rObj = tp.refObject;
var save = () => { rObj.setSavedData(toJSON()); };
var load = () => { fromJSON(rObj.getSavedData()); }


// Basic information
function addBasicInfo(x, y) {
    canvas.addChild(new tp.Text().setText("C"), x, y + 2, 100, 25);
    canvas.addChild(new tp.Text().setText("R"), x, y + 34, 100, 25);
    canvas.addChild(new tp.Text().setText("B"), x, y + 66, 100, 25);
    canvas.addChild(new tp.Text().setText("A"), x, y + 98, 100, 25);
    
    canvas.addChild(classTextBox, x + 15, y, 150, 25);
    canvas.addChild(raceTextBox, x + 15, y + 32, 150, 25);
    canvas.addChild(backgroundTextBox, x + 15, y + 64, 150, 25);

    alignmentSelection.addOption("Lawful Good").addOption("Neutral Good").addOption("Chaotic Good");
    alignmentSelection.addOption("Lawful Neutral").addOption("True Neutral").addOption("Chaotic Neutral");
    alignmentSelection.addOption("Lawful Evil").addOption("Neutral Evil").addOption("Chaotic Evil");
    alignmentSelection.setSelectedIndex(0);
    canvas.addChild(alignmentSelection, x + 15, y + 96, 150, 25);

    canvas.addChild(new tp.Text().setText("Experience"), x, y + 130, 100, 25);
    canvas.addChild(xpTextBox, x + 90, y + 128, 75, 25);
}

canvas.addChild(nameTextBox, 100, 5, 400, 45);
addBasicInfo(5, 60);


// Ability scores
function addAbility(name, x, y) {
    var textBox = addCounter(x + 58, y);
    var modifierText = new tp.Text().setFontSize(19);

    function updateModifierText() {
        modifierText.setText(scoreToModifierString(textBox.getText())); 
    }
    textBox.onTextChanged.add(updateModifierText);

    canvas.addChild(new tp.Text().setText(name).setFontSize(18), x, y + 4, 55, 30);
    canvas.addChild(modifierText, x + 125, y + 4, 50, 30);

    textBox.setText("10");
    return textBox;
}

var y_offset = 0;
for (ability in abilities) {
    abilityTextBoxes[ability] = addAbility(ability, 10, 235 + y_offset);
    y_offset += 50;
}


// Counters for various important quantities
function getProficiency() {
    return parseInt(proficiencyTextBox.getText());
}

function addImportantCounters(x, y) {
    proficiencyTextBox = addCounter(x, y, 2).setText("2");
    canvas.addChild(new tp.Text().setText("Proficiency"), x - 12, y + 43, 100, 20);

    acTextBox = addCounter(x + 105, y).setText("14");
    canvas.addChild(new tp.Text().setText("Armor Class"), x + 88, y + 43, 100, 20);

    initiativeTextBox = addCounter(x, y + 80, 2, -9).setText("0");
    canvas.addChild(new tp.Text().setText("Initiative"), x - 5, y + 123, 100, 20);

    inspirationTextBox = addCounter(x + 105, y + 80).setText("0");
    canvas.addChild(new tp.Text().setText("Inspiration"), x + 93, y + 123, 100, 20);

    currentHPTextBox = addCounter(x - 8, y + 160, 3).setText("10");
    canvas.addChild(new tp.Text().setText("Current HP"), x - 15, y + 203, 100, 20);

    maxHPTextBox = addCounter(x + 97, y + 160, 3).setText("10");
    canvas.addChild(new tp.Text().setText("Maximum HP"), x + 85, y + 203, 105, 20);
}

addImportantCounters(210, 60);


// Death saves
function addDeathSaves(x, y) {
    canvas.addChild(new tp.Text().setText("Death Saves").setFontSize(15), x + 30, y, 200, 30);
    for (cx = x + 15; cx < x + 60; cx += 20) {
        var checkBox = new tp.CheckBox();
        deathSuccessCheckBoxes.push(checkBox);
        canvas.addChild(checkBox, cx, y + 25, 20, 20);
    }
    for (cx = x + 110; cx < x + 160; cx += 20) {
        var checkBox = new tp.CheckBox();
        deathFailureCheckBoxes.push(checkBox);
        canvas.addChild(checkBox, cx, y + 25, 20, 20);
    }
    canvas.addChild(new tp.Text().setText("Successes"), x, y + 40, 100, 20);
    canvas.addChild(new tp.Text().setText("Failures"), x + 105, y + 40, 100, 20);
}

addDeathSaves(205, 295);


// Saves and skills
function addSkill(name, ability, x, y) {
    var checkBox = new tp.CheckBox();
    var modifierText = new tp.Text();
    canvas.addChild(checkBox, x, y, 20, 20);
    canvas.addChild(new tp.Border().setColor(new tp.Color(0.07, 0.07, 0.07)).setChild(modifierText), x + 17, y - 2, 32, 24);
    canvas.addChild(new tp.Text().setText(name), x + 54, y, 200, 25);

    function updateSkillModifier() {
        var modifier = getAbilityModifier(ability);
        if (checkBox.isChecked()) {
            modifier += getProficiency();
        }
        modifierText.setText(modifierToString(modifier));
    };
    checkBox.onCheckStateChanged.add(updateSkillModifier);
    abilityTextBoxes[ability].onTextChanged.add(updateSkillModifier);
    proficiencyTextBox.onTextChanged.add(updateSkillModifier);
    updateSkillModifier();

    return checkBox;
}

var y_offset = 0;
for (ability in abilities) {
    saveCheckBoxes.push(addSkill(abilities[ability], ability, 220, 377 + y_offset));
    y_offset += 26;
}

y_offset = 0;
for (skill in skills) {
    skillCheckBoxes.push(addSkill(skill, skills[skill], 410, 65 + y_offset));
    y_offset += 26;
}


// Add UI to object
var ui = new tp.UIElement();
ui.position = new tp.Vector(3, 0, 0.151);
ui.widget = new tp.Border().setChild(canvas);
ui.width = 600;
ui.height = 537;
ui.scale = 0.5;
ui.useWidgetSize = false;
tp.refObject.attachUI(ui);


// Try to load now, and save every 10 seconds
load();
setInterval(save, 10000);
